@@ -33,10 +33,12 @@
 #define _GNU_SOURCE
 
 #include <ctype.h>
+#include <err.h>
+#include <errno.h>
+#include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <errno.h>
 #include <stdint.h>
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -123,7 +125,7 @@ read_vars(char **namelist,
 	}
 	return;
 err:
-	fprintf(stderr, "efibootmgr: %m\n");
+	warn("efibootmgr");
 	exit(1);
 }
 
@@ -242,16 +244,20 @@ make_boot_var(list_t *boot_list)
 		free_number = opts.bootnum;
 	}
 
-	if (free_number == -1)
+	if (free_number == -1) {
+		warn("efibootmgr: no available boot variables");
 		return NULL;
+	}
 
 	/* Create a new efi_variable_t object
 	   and populate it.
 	*/
 
 	boot = calloc(1, sizeof(*boot));
-	if (!boot)
+	if (!boot) {
+		warn("efibootmgr");
 		return NULL;
+	}
 	if (make_linux_load_option(&boot->data, &boot->data_size) < 0)
 		goto err_boot_entry;
 	if (append_extra_args(&boot->data, &boot->data_size) < 0)
@@ -260,8 +266,10 @@ make_boot_var(list_t *boot_list)
 	boot->num = free_number;
 	boot->guid = EFI_GLOBAL_VARIABLE;
 	rc = asprintf(&boot->name, "Boot%04X", free_number);
-	if (rc < 0)
+	if (rc < 0) {
+		warn("efibootmgr");
 		goto err_boot_entry;
+	}
 	boot->attributes = EFI_VARIABLE_NON_VOLATILE |
 			    EFI_VARIABLE_BOOTSERVICE_ACCESS |
 			    EFI_VARIABLE_RUNTIME_ACCESS;
@@ -272,8 +280,12 @@ make_boot_var(list_t *boot_list)
 	list_add_tail(&boot->list, boot_list);
 	return boot;
 err_boot_entry:
-	if (boot->name)
+	if (boot->name) {
+		warn("Could not set variable %s", boot->name);
 		free(boot->name);
+	} else {
+		warn("Could not set variable");
+	}
 	if (boot->data)
 		free(boot->data);
 	free(boot);
@@ -312,6 +324,15 @@ read_boot_order(efi_variable_t **boot_order)
 	return rc;
 }
 
+static int
+set_boot_u16(const char *name, uint16_t num)
+{
+	return efi_set_variable(EFI_GLOBAL_GUID, name, (uint8_t *)&num,
+				sizeof (num), EFI_VARIABLE_NON_VOLATILE |
+					      EFI_VARIABLE_BOOTSERVICE_ACCESS |
+					      EFI_VARIABLE_RUNTIME_ACCESS);
+}
+
 static int
 add_to_boot_order(uint16_t num)
 {
@@ -321,8 +342,11 @@ add_to_boot_order(uint16_t num)
 	int rc;
 
 	rc = read_boot_order(&boot_order);
-	if (rc < 0)
+	if (rc < 0) {
+		if (errno == ENOENT)
+			rc = set_boot_u16("BootOrder", num);
 		return rc;
+	}
 
 	/* We've now got an array (in boot_order->data) of the
 	 * boot order.  First add our entry, then copy the old array.
@@ -358,8 +382,11 @@ remove_dupes_from_boot_order(void)
 	int rc;
 
 	rc = read_boot_order(&boot_order);
-	if (rc < 0)
+	if (rc < 0) {
+		if (errno == ENOENT)
+			rc = 0;
 		return rc;
+	}
 
 	old_data = (uint16_t *)(boot_order->data);
 	/* Start with the same size */
@@ -409,8 +436,11 @@ remove_from_boot_order(uint16_t num)
 	int rc;
 
 	rc = read_boot_order(&boot_order);
-	if (rc < 0)
+	if (rc < 0) {
+		if (errno == ENOENT)
+			rc = 0;
 		return rc;
+	}
 
 	/* We've now got an array (in boot_order->data) of the
 	   boot order.  Simply copy the array, skipping the
@@ -469,15 +499,6 @@ read_boot_u16(const char *name)
 	return rc;
 }
 
-static int
-set_boot_u16(const char *name, uint16_t num)
-{
-	return efi_set_variable(EFI_GLOBAL_GUID, name, (uint8_t *)&num,
-				sizeof (num), EFI_VARIABLE_NON_VOLATILE |
-					      EFI_VARIABLE_BOOTSERVICE_ACCESS |
-					      EFI_VARIABLE_RUNTIME_ACCESS);
-}
-
 static int
 delete_boot_var(uint16_t num)
 {
@@ -490,13 +511,18 @@ delete_boot_var(uint16_t num)
 	rc = efi_del_variable(EFI_GLOBAL_GUID, name);
 
 	/* For backwards compatibility, try to delete abcdef entries as well */
-	if (rc < 0 && errno == ENOENT) {
-		snprintf(name, sizeof(name), "Boot%04x", num);
-		rc = efi_del_variable(EFI_GLOBAL_GUID, name);
+	if (rc < 0) {
+		if (errno == ENOENT) {
+			snprintf(name, sizeof(name), "Boot%04x", num);
+			rc = efi_del_variable(EFI_GLOBAL_GUID, name);
+		} else if (errno == EPERM) {
+			warn("Could not delete Boot%04X", num);
+			return rc;
+		}
 	}
 
 	if (rc < 0) {
-		fprintf(stderr,"\nboot entry: %X not found\n\n",num);
+		warnx("Boot entry %04X not found", num);
 		return rc;
 	}
 	list_for_each_safe(pos, n, &boot_entry_list) {
@@ -512,7 +538,6 @@ delete_boot_var(uint16_t num)
 	return 0;
 }
 
-
 static void
 set_var_nums(list_t *list)
 {
@@ -570,52 +595,89 @@ is_current_boot_entry(int b)
 	return 0;
 }
 
+static void
+print_error_arrow(char *message, char *buffer, off_t offset)
+{
+	unsigned int i;
+	fprintf(stderr, "%s: %s\n", message, buffer);
+	for (i = 0; i < strlen(message) + 2; i++)
+		fprintf(stderr, " ");
+	for (i = 0; i < offset; i++)
+		fprintf(stderr, " ");
+	fprintf(stderr, "^\n");
+}
 
 static int
 parse_boot_order(char *buffer, uint16_t **order, size_t *length)
 {
-	int i, len;
-	int num, rc;
-
 	uint16_t *data;
 	size_t data_size;
+	size_t len = strlen(buffer);
+	intptr_t end = (intptr_t)buffer + len + 1;
 
-	len = strlen(buffer);
-	if (len % 5 != 4) {
-		fprintf(stderr, "\nInvalid boot order: %s\n\n", buffer);
-		return -1;
+	int num = 0;
+	char *buf = buffer;
+	while ((intptr_t)buf < end) {
+		size_t comma = strcspn(buf, ",");
+		if (comma == 0) {
+			off_t offset = (intptr_t)buf - (intptr_t)buffer;
+			print_error_arrow("Malformed boot order",buffer,offset);
+			exit(8);
+		} else {
+			num++;
+		}
+		buf += comma + 1;
 	}
-	len = (len / 5) + 1;
 
-	data_size = len * sizeof (*data);
-	data = malloc(data_size);
+	data = calloc(num, sizeof (*data));
 	if (!data)
 		return -1;
+	data_size = num * sizeof (*data);
 
-	for (i=0; i < len && *buffer; i++) {
-		rc = sscanf(buffer, "%x", &num);
-		if (rc == 1) {
-			data[i] = num & 0xFFFF;
-		} else {
-			fprintf(stderr,"\nInvalid hex characters in boot order: %s\n\n",buffer);
+	int i = 0;
+	buf = buffer;
+	while ((intptr_t)buf < end) {
+		unsigned long result = 0;
+		size_t comma = strcspn(buf, ",");
+
+		buf[comma] = '\0';
+		char *endptr = NULL;
+		result = strtoul(buf, &endptr, 16);
+		if ((result == ULONG_MAX && errno == ERANGE) ||
+				(endptr && *endptr != '\0')) {
+			print_error_arrow("Invalid boot order", buffer,
+				(intptr_t)endptr - (intptr_t)buffer);
 			free(data);
-			return -1;
+			exit(8);
 		}
+		if (result > 0xffff) {
+			fprintf(stderr, "Invalid boot order entry value: %lX\n",
+				result);
+			print_error_arrow("Invalid boot order", buffer,
+				(intptr_t)buf - (intptr_t)buffer);
+			free(data);
+			exit(8);
+		}
+
 		/* make sure this is an existing boot entry */
-		if (!is_current_boot_entry(data[i])) {
-			fprintf (stderr,"\nboot entry %X does not exist\n\n",data[i]);
+		if (!is_current_boot_entry(result)) {
+			print_error_arrow("Invalid boot order entry value",
+					buffer,
+					(intptr_t)buf - (intptr_t)buffer);
+			fprintf(stderr,"Boot entry %04lX does not exist\n",
+				result);
 			free(data);
-			return -1;
+			exit(8);
 		}
 
-		/* Advance to the comma */ 
-		while (*buffer && *buffer != ',') buffer++;
-		/* Advance through the comma(s) */
-		while (*buffer && *buffer == ',') buffer++;
+		data[i++] = result;
+		buf[comma] = ',';
+		buf += comma + 1;
 	}
+
 	*order = data;
 	*length = data_size;
-	return i;
+	return num;
 }
 
 static int
@@ -1007,15 +1069,26 @@ parse_opts(int argc, char **argv)
 		case 'B':
 			opts.delete_boot = 1;
 			break;
-		case 'b':
-			rc = sscanf(optarg, "%X", &num);
-			if (rc == 1 && num < 0xffff) {
-				opts.bootnum = num;
-			} else {
-				fprintf (stderr,"invalid hex value %s\n",optarg);
+		case 'b': {
+			char *endptr = NULL;
+			unsigned long result;
+			result = strtoul(optarg, &endptr, 16);
+			if ((result == ULONG_MAX && errno == ERANGE) ||
+					(endptr && *endptr != '\0')) {
+				print_error_arrow("Invalid bootnum value",
+					optarg,
+					(intptr_t)endptr - (intptr_t)optarg);
 				exit(1);
 			}
+			if (result > 0xffff) {
+				fprintf(stderr, "Invalid bootnum value: %lX\n",
+					result);
+				exit(1);
+			}
+
+			opts.bootnum = result;
 			break;
+		}
 		case 'c':
 			opts.create = 1;
 			break;
@@ -1050,14 +1123,26 @@ parse_opts(int argc, char **argv)
 			exit(0);
 			break;
 
-		case 'H':
-			rc = sscanf(optarg, "%x", &num);
-			if (rc == 1) opts.acpi_hid = num;
-			else {
-				fprintf (stderr,"invalid hex value %s\n",optarg);
+		case 'H': {
+			char *endptr = NULL;
+			unsigned long result;
+			result = strtoul(optarg, &endptr, 16);
+			if ((result == ULONG_MAX && errno == ERANGE) ||
+					(endptr && *endptr != '\0')) {
+				print_error_arrow("Invalid ACPI_HID value",
+					optarg,
+					(intptr_t)endptr - (intptr_t)optarg);
 				exit(1);
 			}
+			if (result > 0xffff) {
+				fprintf(stderr, "Invalid ACPI_HID value: %lX\n",
+					result);
+				exit(1);
+			}
+
+			opts.acpi_hid = result;
 			break;
+		}
 		case 'i':
 			opts.iface = optarg;
 			break;
@@ -1073,14 +1158,25 @@ parse_opts(int argc, char **argv)
 		case 'N':
 			opts.delete_bootnext = 1;
 			break;
-		case 'n':
-			rc = sscanf(optarg, "%x", &num);
-			if (rc == 1) opts.bootnext = num;
-			else {
-				fprintf (stderr,"invalid hex value %s\n",optarg);
+		case 'n': {
+			char *endptr = NULL;
+			unsigned long result;
+			result = strtoul(optarg, &endptr, 16);
+			if ((result == ULONG_MAX && errno == ERANGE) ||
+					(endptr && *endptr != '\0')) {
+				print_error_arrow("Invalid BootNext value",
+					optarg,
+					(intptr_t)endptr - (intptr_t)optarg);
 				exit(1);
 			}
+			if (result > 0xffff) {
+				fprintf(stderr, "Invalid BootNext value: %lX\n",
+					result);
+				exit(1);
+			}
+			opts.bootnext = result;
 			break;
+		}
 		case 'o':
 			opts.bootorder = optarg;
 			break;
@@ -1116,14 +1212,26 @@ parse_opts(int argc, char **argv)
 			opts.unicode = 1;
 			break;
 
-		case 'U':
-			rc = sscanf(optarg, "%x", &num);
-			if (rc == 1) opts.acpi_uid = num;
-			else {
-				fprintf (stderr,"invalid hex value %s\n",optarg);
+		case 'U': {
+			char *endptr = NULL;
+			unsigned long result;
+			result = strtoul(optarg, &endptr, 16);
+			if ((result == ULONG_MAX && errno == ERANGE) ||
+					(endptr && *endptr != '\0')) {
+				print_error_arrow("Invalid ACPI_UID value",
+					optarg,
+					(intptr_t)endptr - (intptr_t)optarg);
 				exit(1);
 			}
+			if (result > 0xffff) {
+				fprintf(stderr, "Invalid ACPI_UID value: %lX\n",
+					result);
+				exit(1);
+			}
+
+			opts.acpi_uid = result;
 			break;
+		}
 		case 'v':
 			opts.verbose = 1;
 			if (optarg) {
@@ -1177,95 +1285,110 @@ main(int argc, char **argv)
 	if (opts.iface && (
 			opts.acpi_hid < 0 || opts.acpi_uid < 0 ||
 			opts.acpi_hid > UINT32_MAX ||
-			opts.acpi_uid > UINT32_MAX)) {
-		fprintf(stderr, "\nYou must specify the ACPI HID and UID when using -i.\n\n");
-		return 1;
-	}
+			opts.acpi_uid > UINT32_MAX))
+		errx(1, "You must specify the ACPI HID and UID when using -i.");
 
-	if (!efi_variables_supported()) {
-		fprintf(stderr, "\nEFI variables are not supported on this system.\n\n");
-		return 1;
-	}
+	if (!efi_variables_supported())
+		errx(2, "EFI variables are not supported on this system.");
 
 	read_boot_var_names(&boot_names);
 	read_vars(boot_names, &boot_entry_list);
 	set_var_nums(&boot_entry_list);
 
 	if (opts.delete_boot) {
-		if (opts.bootnum == -1) {
-			fprintf(stderr, "\nYou must specify a boot entry to delete (see the -b option).\n\n");
-			return 1;
-		}
-		else
+		if (opts.bootnum == -1)
+			errx(3, "You must specify a boot entry to delete "
+				"(see the -b option).");
+		else {
 			ret = delete_boot_var(opts.bootnum);
+			if (ret < 0)
+				err(15, "Could not delete boot variable");
+		}
 	}
 
 	if (opts.active >= 0) {
-		if (opts.bootnum == -1) {
-			fprintf(stderr, "\nYou must specify a boot entry to activate (see the -b option).\n\n");
-			return 1;
+		if (opts.bootnum == -1)
+			errx(4, "You must specify a boot entry to activate "
+				"(see the -b option");
+		else {
+			ret = set_active_state();
+			if (ret < 0)
+				err(16, "Could not set active state");
 		}
-		else
-			ret=set_active_state();
 	}
 
 	if (opts.create) {
 		warn_duplicate_name(&boot_entry_list);
 		new_boot = make_boot_var(&boot_entry_list);
-		if (!new_boot) {
-			fprintf(stderr, "\nCould not prepare boot variable: %m\n\n");
-			return 1;
-		}
+		if (!new_boot)
+			err(5, "Could not prepare boot variable");
 
 		/* Put this boot var in the right BootOrder */
 		if (new_boot)
 			ret=add_to_boot_order(new_boot->num);
+		if (ret < 0)
+			err(6, "Could not add entry to BootOrder");
 	}
 
 	if (opts.delete_bootorder) {
 		ret = efi_del_variable(EFI_GLOBAL_GUID, "BootOrder");
+		if (ret < 0)
+			err(7, "Could not remove entry from BootOrder");
 	}
 
 	if (opts.bootorder) {
 		ret = set_boot_order(opts.keep_old_entries);
+		if (ret < 0)
+			err(8, "Could not set BootOrder");
 	}
 
 	if (opts.deduplicate) {
 		ret = remove_dupes_from_boot_order();
+		if (ret)
+			err(9, "Could not set BootOrder");
 	}
 
 	if (opts.delete_bootnext) {
+		if (!is_current_boot_entry(opts.delete_bootnext))
+			errx(17, "Boot entry %04X does not exist\n",
+				opts.delete_bootnext);
+
 		ret = efi_del_variable(EFI_GLOBAL_GUID, "BootNext");
+		if (ret < 0)
+			err(10, "Could not delete BootNext");
 	}
 
 	if (opts.delete_timeout) {
 		ret = efi_del_variable(EFI_GLOBAL_GUID, "Timeout");
+		if (ret < 0)
+			err(11, "Could not delete Timeout");
 	}
 
 	if (opts.bootnext >= 0) {
-		if (!is_current_boot_entry(opts.bootnext & 0xFFFF)){
-			fprintf (stderr,"\n\nboot entry %X does not exist\n\n",
-				opts.bootnext);
-			return 1;
-		}
-		ret=set_boot_u16("BootNext", opts.bootnext & 0xFFFF);
+		if (!is_current_boot_entry(opts.bootnext & 0xFFFF))
+			errx(12, "Boot entry %X does not exist", opts.bootnext);
+		ret = set_boot_u16("BootNext", opts.bootnext & 0xFFFF);
+		if (ret < 0)
+			err(13, "Could not set BootNext");
 	}
 
 	if (opts.set_timeout) {
-		ret=set_boot_u16("Timeout", opts.timeout);
+		ret = set_boot_u16("Timeout", opts.timeout);
+		if (ret < 0)
+			err(14, "Could not set Timeout");
 	}
 
 	if (!opts.quiet && ret == 0) {
 		num = read_boot_u16("BootNext");
-		if (num != -1 ) {
+		if (num >= 0) {
 			printf("BootNext: %04X\n", num);
 		}
 		num = read_boot_u16("BootCurrent");
-		if (num != -1) {
+		if (num >= 0) {
 			printf("BootCurrent: %04X\n", num);
 		}
 		num = read_boot_u16("Timeout");
-		if (num != -1) {
+		if (num >= 0) {
 			printf("Timeout: %u seconds\n", num);
 		}
 		show_boot_order();
