@@ -1,8 +1,8 @@
 /*
   scsi_ioctls.[ch]
- 
+
   Copyright (C) 2001 Dell Computer Corporation <Matt_Domsch@dell.com>
- 
+
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
@@ -24,8 +24,15 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <sys/ioctl.h>
+#include <linux/nvme.h>
 #include "scsi_ioctls.h"
 
+int
+get_nvme_ns_id(int fd, uint32_t *ns_id)
+{
+	return ioctl(fd, NVME_IOCTL_ID, &ns_id);
+}
+
 int
 idlun_to_components (Scsi_Idlun *idlun,
 		     unsigned char *host,
@@ -34,7 +41,7 @@ idlun_to_components (Scsi_Idlun *idlun,
 		     unsigned char *lun)
 {
 	if (!idlun || !host || !channel || !id || !lun) return 1;
-	
+
 	*host    = (idlun->dev_id >> 24) & 0xFF;
 	*channel = (idlun->dev_id >> 16) & 0xFF;
 	*id      = (idlun->dev_id      ) & 0xFF;
@@ -43,13 +50,13 @@ idlun_to_components (Scsi_Idlun *idlun,
 }
 
 
-inline int
+int
 get_scsi_idlun(int fd, Scsi_Idlun *idlun)
 {
 	return ioctl(fd, SCSI_IOCTL_GET_IDLUN, idlun);
 }
 
-int 
+int
 get_scsi_pci(int fd, char *slot_name, size_t size)
 {
 	char buf[SLOT_NAME_SIZE] = "";
@@ -81,7 +88,7 @@ int main(int argc, char **argv)
 	memset(&idlun, 0, sizeof(idlun));
 
 	if (argc < 2) {usage(argv); exit(1);}
-	
+
 	fd = open(argv[1], O_RDONLY);
 	if (fd == -1) {
 		perror("Unable to open file");
@@ -96,7 +103,7 @@ int main(int argc, char **argv)
 	if (rc) {
 		perror("Unable to get_scsi_idlun()");
 	}
-	
+
 	printf("Device: %s\n", argv[1]);
 	printf("PCI: %s\n", slot_name);
 
